import copy

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import nn
from torchdiffeq import odeint_adjoint as odeint

from geodiff.transforms import closed_transform_2d, closed_transform_3d


class NeuralODE(nn.Module):
    r"""Neural Ordinary Differential Equation (NeuralODE) architecture as described in [1]_.

    This module generates geometry by evolving a unit circle according a learnable ordinary
    differential equation. The NeuralODE forward and backward passes are handled using the
    `torchdiffeq` library [2]_.

    .. math:
        TODO

    References:
        [1]: Chen, Ricky TQ, et al. "Neural ordinary differential equations." Advances in neural
            information processing systems 31 (2018).
        [2]: Chen, Ricky TQ. "torchdiffeq, 2018." URL https://github.com/rtqichen/torchdiffeq 14.
    """

    def __init__(
        self,
        geometry_dim: int,
        ode_net: nn.Module
    ) -> None:
        r"""Initialize the NeuralODE object.

        Args:
            geometry_dim: Dimension of the output geometry, e.g. 2d, 3d etc.
            ode_net: A torch network for the function defining the neural ordinary differential
                equation.
        """
        super().__init__()

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('geometry_dim', torch.tensor(geometry_dim, dtype = torch.int64))


        # Close the input domain edges to map to the same r, e.g. theta = 0, 2pi in 2D
        if geometry_dim == 2:
            self.closed_transform = closed_transform_2d
        elif geometry_dim == 3:
            self.closed_transform = closed_transform_3d


        # Create an object of ODEFunc class to use with NeuralODE forward pass using the provided
        # ODE network definition
        self.ode_f = ODEFunc(geometry_dim = geometry_dim, ode_net = ode_net)


    def forward(self, T: torch.Tensor = None, num_pts: int = 1000) -> torch.Tensor:
        r"""Compute the coordinates of the shape represented by the NeuralODE object.

        Args:
            T: Options samples in the input domain which are mapped to points on the geometry.
                Mutually exclusive with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on the surface. Mututally
                exclusive with `T`.

        Returns:
            torch.Tensor: Matrix of coordinates of points on the geometry. Shape :math:`(N, d)`,
                where :math:`d` is the dimension of the geometry.
        """
        if T is None:
            if self.geometry_dim == 2:
                T = torch.linspace(0, 1, num_pts).reshape(-1, 1)
            elif self.geometry_dim == 3:
                t, s = torch.meshgrid(torch.linspace(0, 1, num_pts), torch.linspace(0, 1, num_pts),
                                      indexing = 'ij')
                T = torch.stack([t, s], dim = -1).reshape(-1, 2)
        # Put T on the same device as the model
        device = next(self.parameters()).device
        T.to(device = device)

        # Apply the closed transformation
        closed_manifold = self.closed_transform(T)

        # Use closed manifold as initial condition - (N, d)
        y0 = closed_manifold
        # reshape to (N, 1, d) for torchdiffeq batching
        y0 = closed_manifold[:, None, :]

        # Define time integration limits
        time = torch.tensor([0.0, 1.0], device = device)

        Y = odeint(self.ode_f, y0, time, options = {'dtype': torch.float32}).to(device)
        # Get the final shape at time t = 1
        X = Y[-1, :, 0, :]

        return X


    def visualize(self, T: torch.Tensor = None, num_pts: int = 1000, ax = None):
        r"""Plot geometry represented by the NeuralODE object.

        Args:
            T: Options samples in the input domain which are mapped to points on the geometry.
                Mutually exclusive with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on the surface. Mututally
                exclusive with `T`.
            ax: Optional Matplotlib Axes to draw on. If `None`, a new figure/axes is created.

        Returns:
            tuple[fig, ax]: The figure and axes used for plotting.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        # Move to CPU for matplotlib
        X = self.forward(T = T, num_pts = num_pts)
        X = X.detach().cpu()

        # Plot the shape
        ax.plot(X[:, 0], X[:, 1], linestyle = '-', linewidth = 2, color = 'orange', alpha = 0.7,
                label = 'curve')

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        ax.set_title('NeuralODE parameterization')
        ax.legend(loc = 'upper center', bbox_to_anchor = (0.5, -0.12), ncol = 2)

        return fig, ax


class ODEFunc(nn.Module):
    r"""Right-hand side of the ODE: dX/dt = f_θ(X, t).
    """

    def __init__(self, geometry_dim: int, ode_net: nn.Module) -> None:
        r"""Initialize the ODEFunc object.
        
        This forms a thin wrapper around the user-provided `ode_net` to make the forward pass
        compatible with `odeint` function from `torchdiffeq`.

        Args:
            geometry_dim: Dimension of the output geometry, e.g. 2d, 3d etc.
            ode_net: A torch network for the function defining the neural ordinary differential
                equation.
        """
        super().__init__()

        self.geometry_dim = geometry_dim
        self.ode_net = copy.deepcopy(ode_net)


    def forward(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        r"""Compute :math:`dy/dt = f_\theta(y, t)`, where :math:`f_\theta` is the `ode_net`
        defining the ODE evolution.

        Args:
            t: Scalar time at which to compute the time rate of change of state.
            y: State tensor. Shape :math:`(..., d)`, where :math:`d` is the geometry dimension.
        """
        # Broadcast t to appropriate shape to concatenate with y
        t = torch.broadcast_to(t.to(y), y[..., :1].shape)

        # Concatenate y and t to feed as input to our ode function
        state_with_time = torch.cat([y, t], dim = -1)

        # Compute the rate of change of y using the neural network defining the ODE function
        dy = self.ode_net(state_with_time)

        return dy


import torch


def farthest_point_sampling(
    num_pts: int,
    oversampling_factor: int = 16,
    seed: int = 42,
    device: str = 'cpu'
) -> torch.Tensor:
    r"""Greedy Farthest Point Sampling (FPS). Picks exactly `num_pts` samples, blue-noise-like.

    Args:
        num_pts: Total number of points to sample in the domain.
        oversampling_factor: Total points to randomly sample out of which the `num_pts` farthest
            subset will be chosen.
        seed: Random seed to use for generating the points.
        device: The device on which to create the candidate points.
    """

    rng = torch.Generator(device = device).manual_seed(seed)
    # Oversample candidate points randomly to select FPS subset from
    M = num_pts * oversampling_factor
    candidates = torch.rand((M, 2), generator = rng, device = device)

    selected_idx = torch.empty(num_pts, dtype = torch.long, device = device)
    # Start by selecting first candidate as first sample
    selected_idx[0] = 0
    last_selected_pt = candidates[selected_idx[0]]

    # Compute squared distance to nearest selected point so far
    d2 = torch.full((M,), float('inf'), device = device)

    for i in range(1, num_pts):
        # Update nearest-distance-to-set with the last chosen point
        d2 = torch.minimum(d2, torch.sum((candidates - last_selected_pt) ** 2, dim = 1))
        idx = torch.argmax(d2)
        selected_idx[i] = idx
        last_selected_pt = candidates[idx]

    return candidates[selected_idx]


import copy
import math

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F


class NIGnet(nn.Module):
    r"""Neural Injective Geometry network (NIGnet) architecture as described in [1]_.

    This module generates a closed, non-self-intersecting base geometry via a PreAuxNet and then
    applies a stack of NIGnet layers.

    .. math:
        TODO

    References:
        [1]: Atharva Aalok, Juan J. Alonso. 2025. NIGnets. https://github.com/atharvaaalok/NIGnets
    """

    available_intersection_modes = ['possible', 'impossible']

    def __init__(
        self,
        geometry_dim: int,
        layer_count: int,
        preaux_net: nn.Module,
        monotonic_net: nn.Module,
        use_batchnormalization: bool = False,
        use_residual_connection: bool = True,
        intersection_mode: str = 'possible',
    ) -> None:
        r"""Initialize the NIGnet object.

        Args:
            geometry_dim: Dimension of the output geometry, e.g. 2d, 3d etc.
            layer_count: Number of hidden layers in the NICE net.
            preaux_net: A torch network that is used as the Pre-Aux net.
            coupling_net: A torch monotonic network that is used as the monotonic function.
            use_batchnormalization: If `True`, applies batch normalization after each coupling
                layer.
            use_residual_connection: If `True`, applies a residual connection after each coupling
                layer.
            intersection_mode: If `possible`, uses normal Linear layers, else if 'impossible' uses
                ExpLinear layers for guaranteed non-self-intersection.
        """
        super().__init__()

        # Save attributes in buffer so that they can be saved with state_dict
        self.register_buffer('geometry_dim', torch.tensor(geometry_dim, dtype = torch.int64))
        self.register_buffer('layer_count', torch.tensor(layer_count, dtype = torch.int64))
        self.register_buffer('use_batchnormalization', torch.tensor(use_batchnormalization,
                                                               dtype = torch.bool))
        self.register_buffer('use_residual_connection', torch.tensor(use_residual_connection,
                                                               dtype = torch.bool))

        self.intersection_mode = intersection_mode


        # Use a Pre-Aux Net to create baseline closed manifold
        self.preaux_net = copy.deepcopy(preaux_net)


        # Create linear layers and monotonic nets for each NIGnet layer
        linear_class = nn.Linear if intersection_mode == 'possible' else ExpLinear
        self.linear_layers = nn.ModuleList()
        self.monotonic_nets = nn.ModuleList()
        for i in range(layer_count):
            self.linear_layers.append(linear_class(geometry_dim, geometry_dim))
            self.monotonic_nets.append(copy.deepcopy(monotonic_net))


        # Use batchnormalization with each coupling layer
        self.normalization_layers = nn.ModuleList()
        for i in range(layer_count):
            if use_batchnormalization:
                self.normalization_layers.append(nn.BatchNorm1d(geometry_dim))
            else:
                self.normalization_layers.append(nn.Identity())


    def forward(
        self,
        T: torch.Tensor = None,
        num_pts: int = 1000,
        code: torch.Tensor = None,
    ) -> torch.Tensor:
        r"""Compute the coordinates of the shape represented by the NIGnet object.

        Args:
            T: Options samples in the input domain which are mapped to points on the geometry.
                Mutually exclusive with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on the surface. Mututally
                exclusive with `T`.

        Returns:
            torch.Tensor: Matrix of coordinates of points on the geometry. Shape :math:`(N, d)`,
                where :math:`d` is the dimension of the geometry.
        """
        if T is None:
            if self.geometry_dim == 2:
                T = torch.linspace(0, 1, num_pts).reshape(-1, 1)
            elif self.geometry_dim == 3:
                t, s = torch.meshgrid(torch.linspace(0, 1, num_pts), torch.linspace(0, 1, num_pts),
                                      indexing = 'ij')
                T = torch.stack([t, s], dim = -1).reshape(-1, 2)
        # Put T on the same device as the model
        device = next(self.parameters()).device
        T.to(device = device)

        if self.preaux_net.latent_dim != 0:
            # If a single code is provided create copies of it to match the number of T values
            code = code.to(device)
            if T.shape[0] != code.shape[0]:
                code = code.expand(T.shape[0], -1)
            # Otherwise we assume that for each T the right latent vector has been put in code

        # First create a closed transform using the PreAux net
        X = self.preaux_net(T, code)

        for linear_layer, monotonic_net in zip(self.linear_layers, self.monotonic_nets):
            if self.use_residual_connection:
                residual = X

            # Apply linear transformation
            X = linear_layer(X)

            # Apply monotonic network to each component of x separately
            if self.geometry_dim == 2:
                x1, x2 = X[:, 0:1], X[:, 1:2]
                X = torch.cat([monotonic_net(x1), monotonic_net(x2)], dim = 1)
            elif self.geometry_dim == 3:
                x1, x2, x3 = X[:, 0:1], X[:, 1:2], X[:, 2:3]
                X = torch.cat([monotonic_net(x1), monotonic_net(x2), monotonic_net(x3)], dim = 1)

            if self.use_residual_connection:
                # Apply residual connection
                X = (X + residual) / 2.0

        return X


    def visualize(
        self,
        T: torch.Tensor = None,
        num_pts: int = 1000,
        code: torch.Tensor = None,
        ax = None,
    ):
        r"""Plot geometry represented by the NIGnet object.

        Args:
            T: Options samples in the input domain which are mapped to points on the geometry.
                Mutually exclusive with `num_pts`.
            num_pts: Optional number of points :math:`N` to generate on the surface. Mututally
                exclusive with `T`.
            ax: Optional Matplotlib Axes to draw on. If `None`, a new figure/axes is created.

        Returns:
            tuple[fig, ax]: The figure and axes used for plotting.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        # Move to CPU for matplotlib
        X = self.forward(T = T, num_pts = num_pts, code = code)
        X = X.detach().cpu()

        # Plot the shape
        ax.plot(X[:, 0], X[:, 1], linestyle = '-', linewidth = 2, color = 'orange', alpha = 0.7,
                label = 'curve')

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        ax.set_title('NIGnet parameterization')
        ax.legend(loc = 'upper center', bbox_to_anchor = (0.5, -0.12), ncol = 2)

        return fig, ax


class ExpLinear(nn.Module):
    """Linear layer with matrix-exponentiated weight transformation.

    Performs linear transformation using W = exp(weight_matrix) instead of direct weight matrix
    multiplication. This ensures injectivity of the transformation.

    Attributes
    ----------
    in_features : int
        Number of input features.
    out_features : int
        Number of output features.
    W : torch.nn.Parameter
        The learnable weight matrix, which is exponentiated in the forward pass.
    bias : torch.nn.Parameter or None
        Optional learnable bias term for each output feature.

    Parameters
    ----------
    in_features : int
        Number of features in the input.
    out_features : int
        Number of features in the output.
    bias : bool, optional
        If True, a learnable bias is included, by default True.
    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
        """
        Initialize the ExpLinear module.

        Raises
        ------
        AssertionError
            If in_features != out_features.
        """

        super().__init__()

        assert in_features == out_features, 'ExpLinear requires in_features == out_features'

        self.in_features = in_features
        self.out_features = out_features

        self.W = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()


    def reset_parameters(self) -> None:
        """
        Reset the parameters of the layer using Kaiming uniform initialization for the weight
        matrix, and a uniform distribution for the bias.
        """

        nn.init.kaiming_uniform_(self.W, a = math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.W)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Perform the forward pass of ExpLinear.

        The weight matrix `self.W` is exponentiated using `torch.matrix_exp` before being applied to
        the input `x`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (N, in_features).

        Returns
        -------
        torch.Tensor
            Output tensor of shape (N, out_features)
        """

        exp_weight = torch.matrix_exp(self.W)
        return F.linear(x, exp_weight.t(), self.bias)


import torch
import torch.nn as nn
import torch.nn.functional as F


class SmoothMinMaxNet(nn.Module):
    """
    SmoothMinMax Monotonic Network architecture as described in [1]_ that computes a smooth
    "min of max" operation across groups of linear units, with constraints on monotonicity for each
    input dimension. Uses the LogSumExp operation for smooth min and max operations.

    Attributes
    ----------
    available_positivity_constraints : list[str]
        List of valid positivity constraints. Valid values are ['squared', 'exponential']
    input_dim : int
        Number of input features.
    n_groups : int
        Number of groups.
    nodes_per_group : int
        Number of nodes within each group.
    mono_signs : torch.Tensor
        A buffer indicating the monotonic signs for each input dimension of shape (1, 1, input_dim).
        Derived from the 'monotonicity' list.
    positivity_constraint : str
        The method used to ensure positivity of weights. E.g. squaring or exponentiating.
    raw_weights : torch.nn.Parameter
        The raw weights for the linear units of shape (n_groups, nodes_per_group, input_dim).
        These are exponentiated or left unconstrained in `forward()` based on `mono_signs`.
    biases : torch.nn.Parameter
        Bias terms for each node of shape (n_groups, nodes_per_group).
    log_beta : float
        Logarithm of the smooth maximum parameter beta. Beta calculated as exp(log_beta) ensures
        positivity of beta.
    
    Parameters
    ----------
    input_dim : int
        Number of input features.
    n_groups : int
        Number of groups.
    nodes_per_group : int
        Number of nodes within each group.
    monotonicity: list[int], optional
        List specifying monotonicity constraints for each input dimension.
        Possible values for each element:
          - 1  -> strictly increasing
          - -1 -> strictly decreasing
          - 0  -> no monotonic constraint
        If None, a list of all +1 is used (strictly increasing). The default is None.
    init_log_beta : float
        Initial value for log_beta, by default -1.
    positivity_constraint : str
        The method used to ensure positivity of weights. E.g. squaring or exponentiating.
    
    References
    ----------
    .. [1] Igel, C. (2023). Smooth min-max monotonic networks. arXiv preprint arXiv:2306.01147.
    """

    available_positivity_constraints = ['squared', 'exponential']

    def __init__(
        self,
        input_dim: int,
        n_groups: int,
        nodes_per_group: int,
        monotonicity: list = None,
        init_log_beta: float = -1.0,
        positivity_constraint: str = 'squared'
    ) -> None:
        """
        Constructs all the necessary attributes for the MinMaxNet.

        Raises
        ------
        ValueError
            If the length of `monotonicity` is not equal to `input_dim`.
        """

        super().__init__()

        self.input_dim = input_dim
        self.n_groups = n_groups
        self.nodes_per_group = nodes_per_group

        # If no monotonicity info is provided, assume increasing monotonicity for all inputs
        if monotonicity is None:
            monotonicity = [1] * input_dim
        
        if len(monotonicity) != input_dim:
            raise ValueError(
                f'Expected monotonicity to have length {input_dim}, got {len(monotonicity)}.'
            )
        
        if positivity_constraint not in self.available_positivity_constraints:
            raise ValueError(f'Invalid positivity constraint. ' \
                             f'Choose from {self.available_positivity_constraints}')
        self.positivity_constraint = positivity_constraint
        
        # Convert monotonicity from list of -1, 0, 1 to actual signs
        # We store them in a buffer so that pytorch does not treat them as parameters
        self.register_buffer(
            "mono_signs",
            torch.tensor(monotonicity, dtype = torch.float32).view(1, 1, -1)
        )
        
        # raw_weights will be exponentiated in forward() for monotonicity constraints
        self.raw_weights = nn.Parameter(
            torch.empty(n_groups, nodes_per_group, input_dim)
        )
        nn.init.trunc_normal_(self.raw_weights)

        self.biases = nn.Parameter(
            torch.zeros(n_groups, nodes_per_group)
        )
        
        # log_beta = trainable param for LSE "sharpness"
        # beta = exp(log_beta) remains positive
        self.log_beta = nn.Parameter(torch.tensor(init_log_beta, dtype = torch.float32))
    

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Performs the forward pass of the MinMaxNet.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape (batch_size, input_dim).
        
        Returns
        -------
        torch.Tensor
            Output tensor of shape (batch_size, 1).
        """

        # Constrain weights according to monotonicity
        # - If monotonic sign is +1, we use exp(raw_weights) / raw_weights**2 to force positivity
        # - If monotonic sign is -1, we use -exp(raw_weights) / -raw_weights**2 to force negativity
        # - If monotonic sign is 0, we use raw_weights and leave them unconstrained
        if self.positivity_constraint == 'squared':
            w_pos = self.raw_weights ** 2
        elif self.positivity_constraint == 'exponential':
            w_pos = torch.exp(self.raw_weights)
        
        sign_matrix = torch.sign(self.mono_signs).expand_as(self.raw_weights)
        w_actual = torch.where(
            sign_matrix == 0,
            self.raw_weights,
            sign_matrix * w_pos
        )
        
        x_expanded = x.unsqueeze(1).unsqueeze(1)    # (batch_size, 1, 1, input_dim)
        w_expanded = w_actual.unsqueeze(0)          # (1, n_groups, nodes_per_group, input_dim)
        bias_expanded = self.biases.unsqueeze(0)    # (1, n_groups, nodes_per_group)

        lin = (x_expanded * w_expanded).sum(dim = -1)
        lin_out = lin + bias_expanded               # (batch_size, n_groups, nodes_per_group)

        beta = torch.exp(self.log_beta)

        # Within each group, take the max over nodes
        group_out = self._smooth_max(lin_out, beta, dim = 2)    # (batch_size, n_groups)
        # Take the min across groups
        y = -self._smooth_max(-group_out, beta, dim = 1)        # (batch_size,)
        y = y.reshape(-1, 1)

        return y


    @staticmethod
    def _smooth_max(z: torch.Tensor, beta: float, dim: int) -> None:
        r"""
        Computes a smooth approximation to the maximum function using LogSumExp function.

        The smooth maximum with parameter :math:`\beta` is defined as:

        .. math::
            \mathrm{smoothmax}_{\beta}(z) = \frac{1}{\beta} \log
                \left( \sum_i \exp \left( \beta z_i \right) \right)
        
        Parameters
        ----------
        z : torch.Tensor
            Input tensor containing values to compute maximum over.
        beta : float
            Smoothness parameter controlling approximation quality:
        dim : int
            Dimension along which to compute maximum.
        
        Returns
        -------
        torch.Tensor
            Smooth maximum values, reduced along specified dimension.
        """
        return (1.0 / beta) * torch.logsumexp(beta * z, dim = dim)


import torch

from geodiff.sampling import farthest_point_sampling


def sample_T(geometry_dim: int, num_pts: int, device: str = 'cpu'):
    r"""Sample T values in [0, 1]^(d - 1) for d dimensional geometry. These are points that are
    mapped to our final shape.

    Args:
        geometry_dim: Dimension of the final output geometry.
        num_pts: Total points to sample.
        device: The device on which to create the candidate points.
    """

    if geometry_dim == 2:
        T = torch.linspace(0, 1, num_pts, device = device).reshape(-1, 1)
    elif geometry_dim == 3:
        T = farthest_point_sampling(num_pts = num_pts, device = device)

    return T